using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;
using System.ComponentModel.Design;

// --------------------------------------------------------------------
//  To avoid problems produced by the 100%, 125% and 150% DPI settings
//  this project must be compiled with the following flag disabled:
//   "Compatibility" / "Disable Display Scaling On High DPI Settings"
//   (in the file properties of "xxx.exe" and "xxx.vshost.exe")
// --------------------------------------------------------------------

namespace MyPanels
{
    public enum PanelStateOptions
    {
        Collapsed,
        Expanded
    }

    [DesignTimeVisible(true)]
    [Category("Containers")]
    [Description("Visual Studio like Panel")]
    [Designer("System.Windows.Forms.Design.ParentControlDesigner, System.Design", typeof(IDesigner))] 
    public partial class MyPanel : UserControl
    {
        int _expandedHeight;
        PanelStateOptions _panelState = PanelStateOptions.Expanded;
        bool _isCollapsed = false;
        bool _fitToParent = false;
        MyPanel _nextPanel;

        public delegate void DelegateStateChanged();
        [Category("Panels")]
        public event DelegateStateChanged StateChanged;

        [Description("Gets or sets the value for user control height when it is expanded")]
        [DisplayName("Expanded Height")]
        [Category("Panels")]
        [DefaultValueAttribute(0)]
        public int ExpandedHeight
        {
            get { return _expandedHeight; }
            set
            {
                if (value > 0)
                {
                    if (this.DesignMode)
                    {
                        if (_panelState == PanelStateOptions.Expanded)
                        {   //Setting Expanded Height is only allowed when user control is expanded
                            this.SetBounds(this.Location.X,
                                           this.Location.Y,
                                           this.Size.Width,
                                           titlePanel.Height + value);
                        }
                    }
                    else
                    {
                        _expandedHeight = value;
                    }
                }
            }
        }

        [Description("Gets or sets panel title")]
        [DisplayName("Panel Title")]
        [Category("Panels")]
        public string PanelTitle
        {
            get { return lblPanelTitle.Text; }
            set { lblPanelTitle.Text = value; }
        }

        [DefaultValue(typeof(PanelStateOptions),"Expanded")]
        [Description("Gets or sets current panel state")]
        [DisplayName("Panel State")]
        [Category("Panels")]
        public PanelStateOptions PanelState
        {
            get { return _panelState; }
            set 
            { 
                _panelState = value;
                _isCollapsed = (_panelState != PanelStateOptions.Collapsed);
                ToggleState(null,null);
            }
        }

        [Category("Panels")]
        [DesignOnly(true)]
        [DefaultValue(false)]
        [Description("If True, fits the panel to match the parent width")]
        public bool FitToParent
        {
            get { return _fitToParent; }
            set 
            { 
                _fitToParent = value;

                if (_fitToParent)
                {
                    if (this.Parent != null)
                    {
                        this.Location = new Point(0, this.Location.Y);
                        this.Size = new Size(this.Parent.Size.Width, this.Size.Height);
                        this.Anchor = AnchorStyles.Top | AnchorStyles.Left | AnchorStyles.Right;
                    }
                }
                else
                {
                    this.Anchor = AnchorStyles.Top | AnchorStyles.Left;
                }
            }
        }

        [Category("Panels")]
        [Description("Gets or sets the panel to be located beneath this panel")]
        public MyPanel NextPanel
        {
            get { return _nextPanel; }
            set 
            {
                _nextPanel = value;
                MoveNextPanel();
            }
        }

        public MyPanel()
        {
            InitializeComponent();
            this.Load += new EventHandler(Panels_Load);
            this.SizeChanged += new EventHandler(Panels_SizeChanged);
            this.LocationChanged += new EventHandler(Panels_LocationChanged);
        }

        private void Panels_MouseEnter(object sender, EventArgs e)
        {
            //this.Focus();
        }

        void Panels_Load(object sender, EventArgs e)
        {
            _isCollapsed = (_panelState == PanelStateOptions.Collapsed);
        }
        
        void Panels_SizeChanged(object sender, EventArgs e)
        {
            if (this.DesignMode)
            {
                if (_panelState == PanelStateOptions.Expanded)
                {
                    _expandedHeight = this.Height - titlePanel.Height;
                }
                else
                {
                    // Final user can only resize the user control when it is expanded
                    this.SetBounds(this.Location.X,
                                   this.Location.Y,
                                   this.Size.Width,
                                   titlePanel.Height);
                }

                if (this.Parent != null)
                {
                    if (this.Parent.Size.Width != this.Size.Width)
                    {
                        FitToParent = false;
                    }
                }
            }
            MoveNextPanel();
        }

        void Panels_LocationChanged(object sender, EventArgs e)
        {
            if (this.DesignMode)
            {
                if (this.Location.X > 0)
                {
                    FitToParent = false;
                }
            }
            MoveNextPanel();
        }

        private void ToggleState(object sender, EventArgs e)
        {
            if (_isCollapsed)
            {
                // --------------------------------------------------------------------
                //  Here the _expandedHeight is run time recalculated
                //  to overcome problems produced by the 125% or 150% DPI settings
                //  end the presence or not of the flag in the EXE property:
                //    "Compatibility" / "Disable Display Scaling On High DPI Settings" 
                // --------------------------------------------------------------------
                int h = 0;
                for (int i = 0; i < this.Controls.Count; i++)
                {
                    int b = this.Controls[i].Bottom;
                    if (b > h) h = b;
                }
                _expandedHeight = h - titlePanel.Height + 5;

                // ----------------------------------- Expand the user control
                this.SetBounds(this.Location.X,
                                       this.Location.Y,
                                       this.Size.Width,
                                       titlePanel.Height + _expandedHeight);

                // ----------------------------------- Scroll the control completely visibe 
                Panel container = this.Parent.Parent as Panel;
                container.ScrollControlIntoView(this);
            }
            else
            {
                // ----------------------------------- Collapse the user control
                this.SetBounds(this.Location.X,
                                       this.Location.Y,
                                       this.Size.Width,
                                       titlePanel.Height);
            }

            _isCollapsed = !_isCollapsed;

            // Setting content control current state and toggling image
            if (_isCollapsed)
            {
                _panelState = PanelStateOptions.Collapsed;
            }
            else
            {
                _panelState = PanelStateOptions.Expanded;
            }

            if (!this.DesignMode)
            {
                if (StateChanged != null)
                {
                    StateChanged();
                }
            }
        }

        private void MoveNextPanel()
        {
            if (_nextPanel != null)
            {
                _nextPanel.Location = new Point(_nextPanel.Location.X,
                                                this.Location.Y + this.Size.Height);
            }
        }
    }
}