Imports System.Drawing
Imports System.Windows.Forms
Imports MyControls.Properties

Namespace NumericDisplay
	Friend Interface IRenderable
		Sub Render()
	End Interface

	Public Enum EntryMode
		None
		Direct
		Arrow
	End Enum

	Public NotInheritable Class FrequencyEdit
		Inherits UserControl
		Private Const DigitCount As Integer = 10
		Private Const DigitImageSplitCount As Integer = 12
        Private Const DigitSeperatorCount As Integer = (DigitCount - 1) \ 3

		Public Event FrequencyChanged As EventHandler
		Public Event FrequencyChanging As EventHandler(Of FrequencyChangingEventArgs)

		Private ReadOnly _digitControls As FrequencyEditDigit() = New FrequencyEditDigit(DigitCount - 1) {}
		Private ReadOnly _separatorControls As FrequencyEditSeparator() = New FrequencyEditSeparator(DigitSeperatorCount - 1) {}
		Private ReadOnly _imageList As New ImageList()
		Private ReadOnly _digitsImage As Image
		Private ReadOnly _renderTimer As New Timer()
		Private ReadOnly _frequencyChangingEventArgs As New FrequencyChangingEventArgs()
		Private _frequency As Long
		Private _newFrequency As Long
		Private _stepSize As Integer
		Private _editModePosition As Integer
		Private _changingEntryMode As Boolean
		Private _currentEntryMode As EntryMode

		' =====================================================================
		'  Public Properties
		' =====================================================================
		Public Property StepSize() As Integer
			Get
				Return _stepSize
			End Get
			Set
				_stepSize = value
			End Set
		End Property

		Public ReadOnly Property EntryModeActive() As Boolean
			Get
				Return _currentEntryMode <> EntryMode.None
			End Get
		End Property

		Public Property Frequency() As Long
			Get
				Return _frequency
			End Get
			Set
				' TODO ???
				'if (value != _frequency)
				If True Then
					_frequencyChangingEventArgs.Accept = True
					_frequencyChangingEventArgs.Frequency = value
					RaiseEvent FrequencyChanging(Me, _frequencyChangingEventArgs)
					If _frequencyChangingEventArgs.Accept Then
						_frequency = _frequencyChangingEventArgs.Frequency
						UpdateDigitsValues()
						RaiseEvent FrequencyChanged(Me, EventArgs.Empty)
					End If
				End If
			End Set
		End Property

		' =====================================================================
		'  Initialization
		' =====================================================================
		Private digitHeight As Integer
		Private digitWidth As Integer

		Public Sub New()
			' --------------------------------------------------------------------
			'  To avoid problems produced by the 100%, 125% and 150% DPI settings
			'  this project must be compiled with the following flag disabled:
			'   "Compatibility" / "Disable Display Scaling On High DPI Settings"
			'   (in the file properties of "xxx.exe" and "xxx.vshost.exe")
			' --------------------------------------------------------------------

			' ----------------------------------------------------------
			' 11 digits + dot = 11.5
			' ----------------------------------------------------------
			' 18 x 11.5 = 207 
			' 17 x 11.5 = 195.5 
			' 16 x 11.5 = 184 
			' 15 x 11.5 = 172.5 
			' 14 x 11.5 = 161
			' ----------------------------------------------------------
			_digitsImage = Resources.Numbers
			_digitsImage = DirectCast(New Bitmap(_digitsImage, 184, 20), Image)
			digitWidth = 16
			digitHeight = _digitsImage.Height
			SplitDigitImages()

			' ----------------------------------------------------------
			_renderTimer.Interval = 30
			AddHandler _renderTimer.Tick, AddressOf renderTimer_Tick
			_renderTimer.Enabled = True
			DoubleBuffered = True
			ConfigureComponent()
		End Sub

		Private Sub SplitDigitImages()
			_imageList.Images.Clear()
			_imageList.ImageSize = New Size(digitWidth, digitHeight)

			Dim newImage As Bitmap
			Dim xPos As Integer = 0
			For i As Integer = 0 To DigitImageSplitCount - 2
				newImage = New Bitmap(digitWidth, digitHeight)
				Using g As Graphics = Graphics.FromImage(newImage)
					g.DrawImage(_digitsImage, New Rectangle(0, 0, digitWidth, digitHeight), New Rectangle(xPos, 0, digitWidth, digitHeight), GraphicsUnit.Pixel)
				End Using
				xPos += digitWidth

				_imageList.Images.Add(newImage)
			Next

			newImage = New Bitmap(digitWidth, digitHeight)
			Using g As Graphics = Graphics.FromImage(newImage)
				g.DrawImage(_digitsImage, New Rectangle(0, 0, digitWidth, digitHeight), New Rectangle(xPos, 0, digitWidth \ 2, digitHeight), GraphicsUnit.Pixel)
			End Using
			_imageList.Images.Add(newImage)
		End Sub


		Private Sub ConfigureComponent()
			BackColor = Color.Transparent

			If _digitsImage IsNot Nothing Then
				For i As Integer = 0 To DigitCount - 1
					If _digitControls(i) IsNot Nothing AndAlso Controls.Contains(_digitControls(i)) Then
						Controls.Remove(_digitControls(i))
						_digitControls(i) = Nothing
					End If
				Next
				For i As Integer = 0 To DigitSeperatorCount - 1
					If _separatorControls(i) IsNot Nothing AndAlso Controls.Contains(_separatorControls(i)) Then
						Controls.Remove(_separatorControls(i))
						_separatorControls(i) = Nothing
					End If
				Next
			End If

			If _imageList.Images.Count = 0 Then
				Return
			End If

			Dim xPos As Integer = 0
			Dim yPos As Integer = 0

			' -------------------------------------------------- large
			Dim digitWidth As Integer = _imageList.ImageSize.Width
			Dim digitHeight As Integer = _imageList.ImageSize.Height

			For i As Integer = DigitCount - 1 To 0 Step -1
				If (i + 1) Mod 3 = 0 AndAlso i <> DigitCount - 1 Then
					Dim separator As New FrequencyEditSeparator()
					Dim seperatorWidth As Integer = digitWidth \ 2
					Dim seperatorIndex As Integer = i \ 3

					separator.Image = _imageList.Images(11)
					separator.Width = seperatorWidth
					separator.Height = digitHeight
					separator.Location = New Point(xPos, yPos)
					Controls.Add(separator)

					_separatorControls(seperatorIndex) = separator

					xPos += seperatorWidth + 2
				End If

				Dim frequencyEditDigit As New FrequencyEditDigit(i)

				frequencyEditDigit.Location = New Point(xPos, yPos)
				AddHandler frequencyEditDigit.OnDigitClick, AddressOf DigitClickHandler
				AddHandler frequencyEditDigit.MouseLeave, AddressOf DigitMouseLeave

				frequencyEditDigit.Width = digitWidth
				frequencyEditDigit.Height = digitHeight
				frequencyEditDigit.ImageList = _imageList

				Controls.Add(frequencyEditDigit)
				_digitControls(i) = frequencyEditDigit

				xPos += digitWidth + 2
			Next

			Dim weight As Long = 1L
			For i As Integer = 0 To DigitCount - 1
				_digitControls(i).Weight = weight
				weight *= 10
			Next
			Height = digitHeight

			UpdateDigitMask()
		End Sub


		Private Sub renderTimer_Tick(sender As Object, e As EventArgs)
			For i As Integer = 0 To Controls.Count - 1
				If TypeOf Controls(i) Is IRenderable Then
					DirectCast(Controls(i), IRenderable).Render()
				End If
			Next
		End Sub


		' =====================================================================
		'  Handlers
		' =====================================================================
		Private Sub DigitClickHandler(sender As Object, args As FrequencyEditDigitClickEventArgs)

			If _currentEntryMode <> EntryMode.None Then
				LeaveEntryMode()
				Return
			End If

			Dim digit As FrequencyEditDigit = DirectCast(sender, FrequencyEditDigit)
			If digit IsNot Nothing Then
				_newFrequency = _frequency
				If args.Button = MouseButtons.Right Then
					ZeroDigits(digit.DigitIndex)
				Else
					If args.IsUpperHalf AndAlso _frequency >= 0 Then
						IncrementDigit(digit.DigitIndex, True)
					Else
						DecrementDigit(digit.DigitIndex, True)
					End If
				End If

				If _newFrequency <> _frequency Then
					_frequencyChangingEventArgs.Accept = True
					_frequencyChangingEventArgs.Frequency = _newFrequency
					RaiseEvent FrequencyChanging(Me, _frequencyChangingEventArgs)
					If _frequencyChangingEventArgs.Accept Then
						_frequency = _frequencyChangingEventArgs.Frequency
						UpdateDigitsValues()
						RaiseEvent FrequencyChanged(Me, EventArgs.Empty)
					Else
						UpdateDigitsValues()
					End If
				End If
			End If
		End Sub

		Private Sub IncrementDigit(index As Integer, updateDigit As Boolean)
			Dim digit As FrequencyEditDigit = _digitControls(index)
			If digit IsNot Nothing Then
				Dim oldDigit As Integer = digit.DisplayedDigit
				Dim newDigit As Integer = If(digit.DisplayedDigit = 9, 0, digit.DisplayedDigit + 1)

				Dim newFrequency As Long = (_newFrequency - (oldDigit * digit.Weight)) + (newDigit * digit.Weight)

				If updateDigit Then
					digit.DisplayedDigit = newDigit
				End If
				_newFrequency = newFrequency

				If oldDigit = 9 AndAlso index < DigitCount - 1 Then
					IncrementDigit(index + 1, updateDigit)
				End If
			End If
		End Sub

		Private Sub DecrementDigit(index As Integer, updateDigit As Boolean)
			Dim digit As FrequencyEditDigit = _digitControls(index)
			If digit IsNot Nothing Then
				Dim oldDigit As Integer = digit.DisplayedDigit
				Dim newDigit As Integer = If(digit.DisplayedDigit = 0, 9, digit.DisplayedDigit - 1)
				Dim newFrequency As Long = (_newFrequency - (oldDigit * digit.Weight)) + (newDigit * digit.Weight)

				If updateDigit Then
					digit.DisplayedDigit = newDigit
				End If
				_newFrequency = newFrequency

				If oldDigit = 0 AndAlso index < DigitCount - 1 Then
					Dim needDecrement As Boolean = (_newFrequency > Math.Pow(10, index + 1))
					If needDecrement Then
						DecrementDigit(index + 1, updateDigit)
					End If
				End If
			End If
		End Sub

		Private Sub ZeroDigits(index As Integer)
			For i As Integer = 0 To index
				_digitControls(i).DisplayedDigit = 0
			Next
			Dim res As Long = CLng(Math.Truncate(Math.Pow(10, index + 1)))
			_newFrequency = (_newFrequency \ res) * res
		End Sub

		Private Sub UpdateDigitsValues()
			If _digitControls(0) Is Nothing Then
				Return
			End If

			Dim currentFrequency As Long = _frequency

			For i As Integer = DigitCount - 1 To 0 Step -1
				Dim digit As Long = currentFrequency \ _digitControls(i).Weight
				_digitControls(i).DisplayedDigit = CInt(digit)
				currentFrequency -= (_digitControls(i).DisplayedDigit * _digitControls(i).Weight)
			Next

			UpdateDigitMask()
		End Sub

		Private Sub UpdateDigitMask()
			Dim frequency As Long = _frequency

			If frequency >= 0 Then
				For i As Integer = 1 To DigitCount - 1
					If (i + 1) Mod 3 = 0 AndAlso i <> DigitCount - 1 Then
						Dim separatorIndex As Integer = i \ 3
						If _separatorControls(separatorIndex) IsNot Nothing Then
							_separatorControls(separatorIndex).Masked = (_digitControls(i + 1).Weight > frequency)
						End If
					End If
					If _digitControls(i) IsNot Nothing Then
						_digitControls(i).Masked = (_digitControls(i).Weight > frequency)
					End If
				Next
			End If
		End Sub

		Private Sub DigitMouseLeave(sender As Object, e As EventArgs)
			If Not ClientRectangle.Contains(PointToClient(MousePosition)) AndAlso _currentEntryMode <> EntryMode.None Then
				AbortEntryMode()
			End If
		End Sub

		Protected Overrides Sub OnMouseLeave(e As EventArgs)
			MyBase.OnMouseLeave(e)
			If Not ClientRectangle.Contains(PointToClient(MousePosition)) AndAlso _currentEntryMode <> EntryMode.None Then
				AbortEntryMode()
			End If
		End Sub

		Private Function GetFrequencyValue() As Long
			Dim newFrequency As Long = 0
			For i As Integer = 0 To _digitControls.Length - 1
				newFrequency += _digitControls(i).Weight * _digitControls(i).DisplayedDigit
			Next
			Return newFrequency
		End Function

		Private Sub SetFrequencyValue(newFrequency As Long)
			If newFrequency <> _frequency Then
				_frequencyChangingEventArgs.Accept = True
				_frequencyChangingEventArgs.Frequency = newFrequency
				RaiseEvent FrequencyChanging(Me, _frequencyChangingEventArgs)
				If _frequencyChangingEventArgs.Accept Then
					_frequency = _frequencyChangingEventArgs.Frequency
					UpdateDigitsValues()
					RaiseEvent FrequencyChanged(Me, EventArgs.Empty)
				End If
			End If
		End Sub

		' =====================================================================
		'  Keyboard Handling
		' =====================================================================
		Private Sub EnterDirectMode()
			If _changingEntryMode Then
				Return
			End If
			_changingEntryMode = True
			For i As Integer = 0 To _digitControls.Length - 1
				If _digitControls(i) IsNot Nothing Then
					_digitControls(i).Masked = False
					If _digitControls(i).CursorInside Then
						_editModePosition = i
						_digitControls(i).Highlight = True
					End If
				End If
			Next

			ZeroDigits(_digitControls.Length - 1)
			_currentEntryMode = EntryMode.Direct
			_changingEntryMode = False
		End Sub

		Private Sub DirectModeHandler(args As KeyEventArgs)
			Select Case args.KeyCode
				Case Keys.D0, Keys.D1, Keys.D2, Keys.D3, Keys.D4, Keys.D5, _
					Keys.D6, Keys.D7, Keys.D8, Keys.D9, Keys.NumPad0, Keys.NumPad1, _
					Keys.NumPad2, Keys.NumPad3, Keys.NumPad4, Keys.NumPad5, Keys.NumPad6, Keys.NumPad7, _
					Keys.NumPad8, Keys.NumPad9
					Dim newValue As Integer = If((args.KeyCode >= Keys.D0 AndAlso args.KeyCode <= Keys.D9), (args.KeyCode - Keys.D0), (args.KeyCode - Keys.NumPad0))
					_digitControls(_editModePosition).DisplayedDigit = newValue
					If _editModePosition > 0 Then
						_digitControls(_editModePosition).Highlight = False
						_editModePosition -= 1
						_digitControls(_editModePosition).Highlight = True
					Else
						LeaveEntryMode()
					End If
					Exit Select
				Case Keys.Left
					If _editModePosition < _digitControls.Length - 1 Then
						_digitControls(_editModePosition).Highlight = False
						_editModePosition += 1
						_digitControls(_editModePosition).Highlight = True
					End If
					Exit Select
				Case Keys.Right
					If _editModePosition > 0 Then
						_digitControls(_editModePosition).Highlight = False
						_editModePosition -= 1
						_digitControls(_editModePosition).Highlight = True
					End If
					Exit Select
				Case Keys.Back
					_digitControls(_editModePosition).DisplayedDigit = 0
					If _editModePosition < _digitControls.Length - 1 Then
						_digitControls(_editModePosition).Highlight = False
						_editModePosition += 1
						_digitControls(_editModePosition).Highlight = True
					End If
					Exit Select
				Case Keys.Tab, Keys.[Decimal], Keys.OemPeriod
					_digitControls(_editModePosition).Highlight = False
					_editModePosition -= (_editModePosition Mod 3) + 1
					If _editModePosition < 2 Then
						If args.KeyCode = Keys.Tab Then
							_editModePosition = _digitControls.Length - 1
						Else
							_editModePosition = 0
							LeaveEntryMode()
							Exit Select
						End If
					End If
					_digitControls(_editModePosition).Highlight = True
					Exit Select
				Case Keys.Escape
					AbortEntryMode()
					Exit Select
				Case Keys.Enter
					LeaveEntryMode()
					Exit Select
			End Select
		End Sub

		Private Sub EnterArrowMode()
			If _changingEntryMode Then
				Return
			End If
			_changingEntryMode = True
			For i As Integer = 0 To _digitControls.Length - 1
				If _digitControls(i) IsNot Nothing Then
					_digitControls(i).Masked = False
					If _digitControls(i).CursorInside Then
						_editModePosition = i
						_digitControls(i).Highlight = True
					End If
				End If
			Next
			_currentEntryMode = EntryMode.Arrow
			_changingEntryMode = False
		End Sub

		Private Sub ArrowModeHandler(args As KeyEventArgs)
			Select Case args.KeyCode
				Case Keys.Up
					_newFrequency = _frequency
					IncrementDigit(_editModePosition, False)
					SetFrequencyValue(_newFrequency)
					Exit Select
				Case Keys.Down
					_newFrequency = _frequency
					DecrementDigit(_editModePosition, False)
					SetFrequencyValue(_newFrequency)
					Exit Select
				Case Keys.Left
					If _editModePosition < _digitControls.Length - 1 Then
						_digitControls(_editModePosition).Highlight = False
						_editModePosition += 1
						_digitControls(_editModePosition).Highlight = True
					End If
					Exit Select
				Case Keys.Right
					If _editModePosition > 0 Then
						_digitControls(_editModePosition).Highlight = False
						_editModePosition -= 1
						_digitControls(_editModePosition).Highlight = True
					End If
					Exit Select
				Case Keys.Tab
					_digitControls(_editModePosition).Highlight = False
					_editModePosition -= (_editModePosition Mod 3) + 1
					If _editModePosition < 2 Then
						_editModePosition = _digitControls.Length - 1
					End If
					_digitControls(_editModePosition).Highlight = True
					Exit Select
				Case Keys.Enter, Keys.Escape
					AbortEntryMode()
					Exit Select
			End Select
		End Sub


		Private Sub AbortEntryMode()
			If _changingEntryMode Then
				Return
			End If
			_changingEntryMode = True
			_digitControls(_editModePosition).Highlight = False
			UpdateDigitsValues()
			_currentEntryMode = EntryMode.None
			_changingEntryMode = False
		End Sub

		Private Sub LeaveEntryMode()
			If _changingEntryMode Then
				Return
			End If
			_changingEntryMode = True
			_digitControls(_editModePosition).Highlight = False
			If _currentEntryMode = EntryMode.Direct Then
				Dim newFrequency As Long = GetFrequencyValue()
				SetFrequencyValue(newFrequency)
			End If
			_currentEntryMode = EntryMode.None
			_changingEntryMode = False
		End Sub

		Private Function DigitKeyHandler(args As KeyEventArgs) As Boolean
			If Not ClientRectangle.Contains(PointToClient(MousePosition)) OrElse _changingEntryMode Then
				Return False
			End If

			If _currentEntryMode <> EntryMode.None Then
				Select Case _currentEntryMode
					Case EntryMode.Direct
						DirectModeHandler(args)
						Exit Select
					Case EntryMode.Arrow
						ArrowModeHandler(args)
						Exit Select
				End Select
				Return True
			End If

			If (args.KeyCode >= Keys.D0 AndAlso args.KeyCode <= Keys.D9) OrElse (args.KeyCode >= Keys.NumPad0 AndAlso args.KeyCode <= Keys.NumPad9) Then

				EnterDirectMode()
				DirectModeHandler(args)
				Return True
			End If

			If args.KeyCode = Keys.Up OrElse args.KeyCode = Keys.Down OrElse args.KeyCode = Keys.Left OrElse args.KeyCode = Keys.Right Then
				EnterArrowMode()
				ArrowModeHandler(args)
				Return True
			End If

			If args.Modifiers = Keys.Control Then
				If args.KeyCode = Keys.C Then
					Dim frequency As String = String.Format("{0}", GetFrequencyValue())
					Clipboard.SetText(frequency, TextDataFormat.Text)
					Return True
				End If
				If args.KeyCode = Keys.V Then
					Dim newFrequency As Long = 0
					Dim result As Boolean = Long.TryParse(Clipboard.GetText(), newFrequency)
					If result Then
						SetFrequencyValue(newFrequency)
					End If
					Return True
				End If
			End If

			Return False
		End Function

		Protected Overrides Function ProcessCmdKey(ByRef msg As Message, keyData As Keys) As Boolean
			Const  WM_KEYDOWN As Integer = &H100
			Const  WM_SYSKEYDOWN As Integer = &H104

			If (msg.Msg = WM_KEYDOWN) OrElse (msg.Msg = WM_SYSKEYDOWN) Then
				Return DigitKeyHandler(New KeyEventArgs(keyData))
			End If
			Return MyBase.ProcessCmdKey(msg, keyData)
		End Function
	End Class

	Public Class FrequencyChangingEventArgs
		Inherits EventArgs
		Public Property Frequency() As Long
			Get
				Return m_Frequency
			End Get
			Set
				m_Frequency = Value
			End Set
		End Property
		Private m_Frequency As Long
		Public Property Accept() As Boolean
			Get
				Return m_Accept
			End Get
			Set
				m_Accept = Value
			End Set
		End Property
		Private m_Accept As Boolean
	End Class
End Namespace
