Imports System.Drawing
Imports System.Drawing.Imaging
Imports System.Windows.Forms

Namespace NumericDisplay
	Public Delegate Sub OnDigitClickDelegate(sender As Object, args As FrequencyEditDigitClickEventArgs)

	Friend NotInheritable Class FrequencyEditDigit
		Inherits UserControl
		Implements IRenderable
		Private Const MaskedDigitTransparency As Single = 0.3F

		Public Event OnDigitClick As OnDigitClickDelegate

		Private _masked As Boolean
		Private _displayedDigit As Integer
		Private _weight As Long
		Private _renderNeeded As Boolean
		Private _cursorInside As Boolean
		Private _highlight As Boolean
		Private _lastMouseY As Integer
		Private _lastIsUpperHalf As Boolean
		Private _isUpperHalf As Boolean
		Private _tickCount As Integer
		Private _imageList As ImageList
		Private ReadOnly _digitIndex As Integer
		Private ReadOnly _tickTimer As New Timer()
		Private ReadOnly _maskedAttributes As New ImageAttributes()

		Public Property ImageList() As ImageList
			Get
				Return _imageList
			End Get
			Set
				_imageList = value
			End Set
		End Property

		Public Property Highlight() As Boolean
			Get
				Return _highlight
			End Get
			Set
				_highlight = value
				_renderNeeded = True
			End Set
		End Property

		Public ReadOnly Property CursorInside() As Boolean
			Get
				Return _cursorInside
			End Get
		End Property

		Public Property DisplayedDigit() As Integer
			Get
				Return _displayedDigit
			End Get
			Set
				If value >= 0 AndAlso value <= 9 Then
					If _displayedDigit <> value Then
						_displayedDigit = value
						_renderNeeded = True
					End If
				End If
			End Set
		End Property

		Public ReadOnly Property DigitIndex() As Integer
			Get
				Return _digitIndex
			End Get
		End Property

		Public Property Masked() As Boolean
			Get
				Return _masked
			End Get
			Set
				If _masked <> value Then
					_masked = value
					_renderNeeded = True
				End If
			End Set
		End Property

		Public Property Weight() As Long
			Get
				Return _weight
			End Get
			Set
				_weight = value
			End Set
		End Property

		Public Sub New(digitIndex As Integer)
			DoubleBuffered = True
			BackColor = Color.Transparent
			AddHandler _tickTimer.Tick, AddressOf timer_Tick
			UpdateStyles()
			_digitIndex = digitIndex

			Dim cm As New ColorMatrix()
			cm.Matrix33 = MaskedDigitTransparency
			_maskedAttributes.SetColorMatrix(cm, ColorMatrixFlag.[Default], ColorAdjustType.Bitmap)
		End Sub

		Protected Overrides Sub OnPaint(e As PaintEventArgs)
			If _imageList IsNot Nothing Then
				If _displayedDigit < _imageList.Images.Count Then
					Dim _image As Image = _imageList.Images(_displayedDigit)
					Dim attributes As ImageAttributes = If(((_masked AndAlso Not _cursorInside) OrElse Not Parent.Enabled), _maskedAttributes, Nothing)

					e.Graphics.DrawImage(_image, New Rectangle(0, 0, _image.Width, _image.Height), 0F, 0F, _image.Width, _image.Height, _
						GraphicsUnit.Pixel, attributes)
				End If
			End If

			If _cursorInside AndAlso Not DirectCast(Parent, FrequencyEdit).EntryModeActive Then
				Dim isUpperHalf As Boolean = (_lastMouseY <= ClientRectangle.Height \ 2)

				Dim transparentColor As Color = Color.FromArgb(100, If(isUpperHalf, Color.Red, Color.Blue))
				Using transparentBrush As New SolidBrush(transparentColor)
					If isUpperHalf Then
						e.Graphics.FillRectangle(transparentBrush, New Rectangle(0, 0, ClientRectangle.Width, ClientRectangle.Height \ 2))
					Else
						e.Graphics.FillRectangle(transparentBrush, New Rectangle(0, ClientRectangle.Height \ 2, ClientRectangle.Width, ClientRectangle.Height))
					End If
				End Using
			End If

			If _highlight Then
				Dim transparentColor As New SolidBrush(Color.FromArgb(25, Color.Red))
				e.Graphics.FillRectangle(transparentColor, New Rectangle(0, 0, ClientRectangle.Width, ClientRectangle.Height))
			End If
		End Sub

		Protected Overrides Sub OnMouseMove(e As MouseEventArgs)
			MyBase.OnMouseMove(e)

			_isUpperHalf = (e.Y <= ClientRectangle.Height \ 2)

			_lastMouseY = e.Y

			If _isUpperHalf <> _lastIsUpperHalf Then
				_renderNeeded = True
				_tickCount = 0
			End If

			_lastIsUpperHalf = _isUpperHalf
		End Sub

		Protected Overrides Sub OnMouseEnter(e As EventArgs)
			MyBase.OnMouseEnter(e)
			_cursorInside = True
			_renderNeeded = True
			Focus()
		End Sub

		Protected Overrides Sub OnMouseLeave(e As EventArgs)
			MyBase.OnMouseLeave(e)
			_cursorInside = False
			_renderNeeded = True
		End Sub

		Protected Overrides Sub OnMouseDown(e As MouseEventArgs)
			MyBase.OnMouseDown(e)
			_isUpperHalf = (e.Y <= ClientRectangle.Height \ 2)
			RaiseEvent OnDigitClick(Me, New FrequencyEditDigitClickEventArgs(_isUpperHalf, e.Button))
			_tickCount = 1
			_tickTimer.Interval = 300
			_tickTimer.Enabled = True
		End Sub

		Protected Overrides Sub OnMouseUp(e As MouseEventArgs)
			_tickTimer.Enabled = False
		End Sub

		Protected Overrides Sub OnMouseWheel(e As MouseEventArgs)
			MyBase.OnMouseWheel(e)
			If Not Me.ClientRectangle.Contains(e.Location) Then
				Return
			End If
			RaiseEvent OnDigitClick(Me, New FrequencyEditDigitClickEventArgs((e.Delta > 0), e.Button))
		End Sub

		Public Sub Render() Implements IRenderable.Render
			If _renderNeeded Then
				Invalidate()
				_renderNeeded = False
			End If
		End Sub

		Private Sub timer_Tick(sender As Object, e As EventArgs)
			RaiseEvent OnDigitClick(Me, New FrequencyEditDigitClickEventArgs(_isUpperHalf, MouseButtons.Left))
			_tickCount += 1
			Select Case _tickCount
				Case 4
					_tickTimer.Interval = 100
					Exit Select
				Case 8
					_tickTimer.Interval = 50
					Exit Select
				Case 12
					_tickTimer.Interval = 20
					Exit Select
				Case 20
					_tickTimer.Interval = 10
					Exit Select
			End Select
		End Sub
	End Class

	Public Class FrequencyEditDigitClickEventArgs
		Public IsUpperHalf As Boolean
		Public Button As MouseButtons

		Public Sub New(isUpperHalf__1 As Boolean, button__2 As MouseButtons)
			IsUpperHalf = isUpperHalf__1
			Button = button__2
		End Sub
	End Class
End Namespace
